import unittest

from filtrark.filtrark import Filtrark


class TestFiltrark(unittest.TestCase):
    """Tests for `filtrark` package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        self.filtrark = Filtrark()

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_filtrark_object_creation(self):
        """Test something."""
        self.assertTrue(isinstance(self.filtrark, Filtrark))

    def test_filtrark_parse_tuple(self):
        filter_tuple = ('field', '=', 99)
        expected = 'field = 99'
        result = self.filtrark._parse_term(filter_tuple)
        self.assertEqual(result, expected)

    def test_filtrark_parse_single_term(self):
        domain = [('field', '=', 7)]
        expected = 'field = 7'
        result = self.filtrark.parse(domain)
        self.assertEqual(result, expected)

    def test_filtrark_match_term_operator(self):
        operator_dict = {'=': '=', '!=': '<>', '<=': '<=', '<': '<',
                         '>': '>', '>=': '>=', 'in': 'in'}
        for key, value in operator_dict.items():
            result = self.filtrark._match_term_operator(key)
            self.assertEqual(result, value)

    def test_filtrark_default_join(self):
        stack = ['field2 <> 8', 'field = 7']
        expected = 'field = 7 AND field2 <> 8'
        result = self.filtrark._default_join(stack)
        self.assertEqual(result, [expected])

    def test_filtrark_parse_multiple_terms(self):
        test_domains = [
            ([('field', '=', 7), ('field2', '!=', 8)],
             'field = 7 AND field2 <> 8'),
            ([('field', '=', 7), ('field2', '!=', 8), ('field3', '>=', 9)],
             'field = 7 AND field2 <> 8 AND field3 >= 9'),
            (['|', ('field', '=', 7), ('field2', '!=', 8), ('field3', '>=', 9)],
             'field = 7 OR field2 <> 8 AND field3 >= 9'),
            (['|', ('field', '=', 7),
              '!', ('field2', '!=', 8), ('field3', '>=', 9)],
             'field = 7 OR NOT field2 <> 8 AND field3 >= 9'),
            (['!', ('field', '=', 7)], 'NOT field = 7'),
        ]

        for test_domain in test_domains:
            result = self.filtrark.parse(test_domain[0])
            expected = test_domain[1]
            self.assertEqual(result, expected)



from typing import NewType, List, Tuple, Union

TermTuple = NewType('TermTuple', Tuple[str, str, Union[str, float]])


class Filtrark:
    def __init__(self) -> None:
        self.term_operators = {
            '=': '=', '!=': '<>', '<=': '<=', '<': '<',
            '>': '>', '>=': '>=', 'in': 'in'}

        self.binary_operators = {
            '&': lambda a, b: a + ' AND ' + b,
            '|': lambda a, b: a + ' OR ' + b}

        self.unary_operators = {
            '!': lambda a: 'NOT ' + a}

    def parse(self, domain: List[Union[str, TermTuple]]) -> str:
        stack = []
        for item in list(reversed(domain)):
            if item in self.binary_operators:
                first_operand = stack.pop()
                second_operand = stack.pop()
                stack.append(self.binary_operators[item](
                    first_operand, second_operand))
            elif item in self.unary_operators:
                operand = stack.pop()
                stack.append(self.unary_operators[item](operand))

            stack = self._default_join(stack)

            if isinstance(item, tuple):
                result = self._parse_term(item)
                stack.append(result)

        result = self._default_join(stack)[0]
        return result

    def _default_join(self, stack: List[Union[str, TermTuple]]
                      ) -> List[Union[str, TermTuple]]:
        if len(stack) == 2:
            first_operand = stack.pop()
            second_operand = stack.pop()
            stack.append(self.binary_operators['&'](
                first_operand, second_operand))
        return stack

    def _match_term_operator(self, operator: str) -> str:
        return self.term_operators.get(operator)

    def _parse_term(self, term_tuple: TermTuple) -> str:
        field, operator, value = term_tuple
        operator = self._match_term_operator(operator)
        result = "{} {} {}".format(field, operator, value)
        return result
